@@ -1,7 +1,7 @@
 import json
 from enum import Enum
 import requests
-from typing import Optional, List
+from typing import Optional, List, Union
 
 
 class SolanaMintAddresses:
@@ -42,9 +42,92 @@ class SearchMethod(Enum):
     EXACT_MATCH = "exact_match"
 
 
+class DerivationPath(Enum):
+
+    CLI_PATH = ""
+    PHANTOM_AND_SOLLET_PATH = "m/44/501/0/0"
+    SOLFLARE_PATH = "m/44/501/0"
+
+    @staticmethod
+    def get_phantom_wallet_derivation_path(wallet_index: int) -> str:
+        try:
+            assert isinstance(wallet_index, int)
+            assert wallet_index >= 0
+        except AssertionError:
+            raise Exception("Wallet index must be an integer greater than or equal to 0.")
+        return f"m/44/501/{wallet_index}/0"
+
+
+class SolanaWallet:
+
+    def __init__(
+        self,
+        secret_recovery_phrase: str = None,
+        private_key: List[int] = None,
+        b58_private_key: str = None,
+        derivation_path: Union[str, DerivationPath] = DerivationPath.PHANTOM_AND_SOLLET_PATH,
+        passphrase: str = str()
+    ):
+        supplied = int(secret_recovery_phrase is not None) + int(private_key is not None) \
+                   + int(b58_private_key is not None)
+        if supplied != 1:
+            raise Exception("Provide EXACTLY ONE of `secret_recovery_phrase` OR `private_key` OR `b58_private_key`.")
+        if secret_recovery_phrase is not None and not isinstance(secret_recovery_phrase, str):
+            raise Exception("`secret_recovery_phrase` must be a `str`.")
+        if private_key is not None and not isinstance(private_key, list):
+            raise Exception(
+                "`private_key` must be a `list`. Example: [99, 110, 111, ..., 88, 88, 17]. "
+                "If you are trying to provide a `str` private key from Phantom, use the argument "
+                "`b58_private_key`"
+            )
+        if b58_private_key is not None and not isinstance(b58_private_key, str):
+            raise Exception("`b58_private_key` must be a `str`.")
+
+        self.secret_recovery_phrase = secret_recovery_phrase
+        self.private_key = private_key
+        self.b58_private_key = b58_private_key
+
+        if secret_recovery_phrase is not None:
+            if isinstance(derivation_path, DerivationPath):
+                derivation_path = derivation_path.value
+            elif not isinstance(derivation_path, str):
+                raise Exception("`derivation_path` must be a `str` or instance of the enum `DerivationPath`.")
+
+            if not isinstance(passphrase, str):
+                raise Exception("`passphrase` must be a `str`.")
+
+        self.derivation_path = derivation_path
+        self.passphrase = passphrase
+
+    def get_formatted_request_payload(self) -> dict:
+        if self.secret_recovery_phrase is not None:
+            return {
+                'wallet': {
+                    'secret_recovery_phrase': self.secret_recovery_phrase,
+                    'derivation_path': self.derivation_path,
+                    'passphrase': self.passphrase
+                }
+            }
+        elif self.private_key is not None:
+            return {
+                'wallet': {
+                    'private_key': self.private_key
+                }
+            }
+        elif self.b58_private_key is not None:
+            return {
+                'wallet': {
+                    'b58_private_key': self.b58_private_key
+                }
+            }
+        else:
+            raise Exception("Unknown error. Improperly initialized instance of `SolanaWallet`.")
+
+
 class TheBlockchainAPIResource:
 
     __url = "https://api.theblockchainapi.com/v1/"
+    __timeout = 120
 
     class __RequestMethod(Enum):
         GET = "GET"
@@ -89,13 +172,18 @@ class TheBlockchainAPIResource:
         if headers is None:
             headers = self._get_headers()
 
+        if len(payload) > 0:
+            payload = json.dumps(payload)
+        else:
+            payload = None
+
         if files is None:
             r = requests.request(
                 request_method.value,
                 url=self.__url + endpoint,
-                params=payload,
+                data=payload,
                 headers=headers,
-                timeout=120
+                timeout=self.__timeout
             )
         else:
             r = requests.request(
@@ -104,7 +192,7 @@ class TheBlockchainAPIResource:
                 data=payload,
                 files=files,
                 headers=headers,
-                timeout=120
+                timeout=self.__timeout
             )
         try:
             json_content = json.loads(r.content)
@@ -133,37 +221,39 @@ class TheBlockchainAPIResource:
         """
         response = self._request(
             payload=dict(),
-            endpoint="solana/wallet/secret_recovery_phrase",
+            endpoint="solana/wallet/generate/secret_recovery_phrase",
             request_method=self.__RequestMethod.POST
         )
         if 'error_message' in response:
             raise Exception(response['error_message'])
         return response['secret_recovery_phrase']
 
-    def derive_public_key(
-        self,
-        secret_recovery_phrase: str,
-        derivation_path: Optional[str] = None,
-        passphrase: str = str()
-    ) -> str:
+    def generate_private_key(self) -> str:
+        """
+        More info:
+        https://docs.theblockchainapi.com/#operation/solanaGeneratePrivateKey
+        :return:
+        """
+        response = self._request(
+            payload=dict(),
+            endpoint="solana/wallet/generate/private_key",
+            request_method=self.__RequestMethod.POST
+        )
+        if 'error_message' in response:
+            raise Exception(response['error_message'])
+        return response
+
+    def derive_public_key(self, wallet: SolanaWallet) -> str:
         """
         Derives a public key given the info.
         More info:
         https://docs.theblockchainapi.com/#operation/solanaDerivePublicKey
-        :param secret_recovery_phrase:
-        :param derivation_path: Derivation path default matches the CLI. Use "m/44/501/0/0" to match Phantom.
-        :param passphrase:
+        :param wallet:
         :return:
         """
-        payload = {
-            "secret_recovery_phrase": secret_recovery_phrase,
-            "passphrase": passphrase
-        }
-        if derivation_path is not None:
-            payload["derivation_path"] = derivation_path
 
         response = self._request(
-            payload=payload,
+            payload=wallet.get_formatted_request_payload(),
             endpoint="solana/wallet/public_key",
             request_method=self.__RequestMethod.POST
         )
@@ -171,6 +261,21 @@ class TheBlockchainAPIResource:
             raise Exception(response['error_message'])
         return response['public_key']
 
+    def derive_private_key(self) -> str:
+        """
+        More info:
+        https://docs.theblockchainapi.com/#operation/solanaDerivePrivateKey
+        :return:
+        """
+        response = self._request(
+            payload=dict(),
+            endpoint="solana/wallet/private_key",
+            request_method=self.__RequestMethod.POST
+        )
+        if 'error_message' in response:
+            raise Exception(response['error_message'])
+        return response
+
     def get_balance(
         self,
         public_key: str,
@@ -338,35 +443,26 @@ class TheBlockchainAPIResource:
 
     def transfer(
         self,
-        secret_recovery_phrase: str,
+        wallet: SolanaWallet,
         recipient_address: str,
         token_address: Optional[str] = None,
-        derivation_path: Optional[str] = None,
-        passphrase: str = str(),
         network: SolanaNetwork = SolanaNetwork.DEVNET,
         amount: str = "1"
     ) -> str:
         """
         More info:
         https://docs.theblockchainapi.com/#operation/solanaTransfer
-        :param secret_recovery_phrase:
+        :param wallet:
         :param recipient_address:
         :param token_address: If not provided, defaults to transferring SOL
-        :param derivation_path: Derivation path default matches the CLI. Use "m/44/501/0/0" to match Phantom.
-        :param passphrase:
         :param network:
         :param amount:
         :return:
         """
-        payload = {
-            "recipient_address": recipient_address,
-            "secret_recovery_phrase": secret_recovery_phrase,
-            "passphrase": passphrase,
-            "network": network.value,
-            "amount": amount
-        }
-        if derivation_path is not None:
-            payload['derivation_path'] = derivation_path
+        payload = wallet.get_formatted_request_payload()
+        payload["network"] = network.value
+        payload["amount"] = amount
+        payload["recipient_address"] = recipient_address
         if token_address is not None:
             payload["token_address"] = token_address
 
@@ -381,10 +477,9 @@ class TheBlockchainAPIResource:
 
     def create_nft(
         self,
-        secret_recovery_phrase: str,
-        derivation_path: Optional[str] = None,
-        passphrase: str = str(),
+        wallet: SolanaWallet,
         network: SolanaNetwork = SolanaNetwork.DEVNET,
+        mint_to_public_key: str = None,
         nft_name: str = str(),
         nft_symbol: str = str(),
         nft_description: str = str(),
@@ -400,10 +495,9 @@ class TheBlockchainAPIResource:
         """
         More info:
         https://docs.theblockchainapi.com/#operation/solanaCreateNFT
-        :param secret_recovery_phrase:
-        :param derivation_path: Derivation path default matches the CLI. Use "m/44/501/0/0" to match Phantom.
-        :param passphrase:
+        :param wallet:
         :param network:
+        :param mint_to_public_key: Assign ownership of the NFT after minting it
         :param nft_name: The name of the NFT
         :param nft_symbol: The symbol of the NFT
         :param nft_description: The description of the NFT
@@ -420,9 +514,8 @@ class TheBlockchainAPIResource:
         """
         if nft_metadata is None:
             nft_metadata = dict()
+        wallet_payload = wallet.get_formatted_request_payload()
         payload = {
-            "secret_recovery_phrase": secret_recovery_phrase,
-            "passphrase": passphrase,
             "network": network.value,
             "nft_name": nft_name,
             "nft_symbol": nft_symbol,
@@ -434,12 +527,13 @@ class TheBlockchainAPIResource:
             "is_master_edition": is_master_edition,
             "seller_fee_basis_points": seller_fee_basis_points
         }
-        if derivation_path is not None:
-            payload['derivation_path'] = derivation_path
+        payload = {**payload, **wallet_payload}
         if creators is not None:
-            payload['creators'] = json.dumps(creators)
+            payload['creators'] = creators
         if share is not None:
-            payload['share'] = json.dumps(share)
+            payload['share'] = share
+        if mint_to_public_key is not None:
+            payload['mint_to_public_key'] = mint_to_public_key
 
         response = self._request(
             payload=payload,
@@ -592,9 +686,7 @@ class TheBlockchainAPIResource:
     def mint_from_candy_machine(
         self,
         config_address: str,
-        secret_recovery_phrase: str,
-        derivation_path: Optional[str] = None,
-        passphrase: str = str(),
+        wallet: SolanaWallet,
         candy_machine_contract_version: SolanaCandyMachineContractVersion = SolanaCandyMachineContractVersion.V1,
         network: SolanaNetwork = SolanaNetwork.DEVNET
     ):
@@ -605,23 +697,19 @@ class TheBlockchainAPIResource:
         You can retrieve this if you have the candy machine ID using
         this endpoint (https://docs.theblockchainapi.com/#operation/solanaGetCandyMachineDetails)
         and retrieving the config_address from the response..
-        :param secret_recovery_phrase:
-        :param derivation_path:
-        :param passphrase:
+        :param wallet:
         :param candy_machine_contract_version:
         :param network:
         :return: A task_id. Use the `get_task` function to retrieve the result once this task has completed processing.
         You can poll the `get_task` function to see results.
         """
+        wallet_payload = wallet.get_formatted_request_payload()
         payload = {
-            "secret_recovery_phrase": secret_recovery_phrase,
             "network": network.value,
-            "passphrase": passphrase,
             "config_address": config_address,
             "candy_machine_contract_version": candy_machine_contract_version.value
         }
-        if derivation_path is not None:
-            payload['derivation_path'] = derivation_path
+        payload = {**payload, **wallet_payload}
         response = self._request(
             payload=payload,
             endpoint="solana/nft/candy_machine/mint",
@@ -692,9 +780,7 @@ class TheBlockchainAPIResource:
 
     def create_test_candy_machine(
         self,
-        secret_recovery_phrase: str,
-        derivation_path: Optional[str] = None,
-        passphrase: str = str(),
+        wallet: SolanaWallet,
         include_gatekeeper: bool = False,
         network: SolanaNetwork = SolanaNetwork.DEVNET,
         candy_machine_contract_version: SolanaCandyMachineContractVersion = SolanaCandyMachineContractVersion.V1,
@@ -702,22 +788,19 @@ class TheBlockchainAPIResource:
         """
         Mint Info:
         https://docs.theblockchainapi.com/#operation/solanaCreateTestCandyMachine
-        :param secret_recovery_phrase:
-        :param derivation_path:
-        :param passphrase:
+        :param wallet:
         :param network:
+        :param include_gatekeeper:
         :param candy_machine_contract_version:
         :return:
         """
+        wallet_payload = wallet.get_formatted_request_payload()
         payload = {
-            "secret_recovery_phrase": secret_recovery_phrase,
             "network": network.value,
-            "passphrase": passphrase,
             "candy_machine_contract_version": candy_machine_contract_version.value,
             "include_gatekeeper": include_gatekeeper
         }
-        if derivation_path is not None:
-            payload['derivation_path'] = derivation_path
+        payload = {**payload, **wallet_payload}
         response = self._request(
             payload=payload,
             endpoint="solana/nft/candy_machine",
